#!/usr/bin/env python3

import sys
import filecmp
import os
import unittest
from pyfastaq import sequences, utils, intervals, tasks

modules_dir = os.path.dirname(os.path.abspath(sequences.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class Error (Exception): pass

expected_embl = [
    'aaacaaaccaaatatggattttattgtagccatatttgctctgtttgttattagctcattcacaattacttccacaaatgcagttgaagcttctactcttcttgacataggtaacctgagtcggagcagttttcctcgtggcttcatctttggtgctggatcttcagcataccaatttgaaggtgcagtaaacgaaggcggtagaggaccaagtatttgggataccttcacccataaatatccagaaaaaataagggatggaagcaatgcagacatcacggttgaccaatatcaccgctacaaggaagatgttgggattatgaaggatcaaaatatggattcgtatagattctcaatctcttggccaagaatactcccaaagggaaagttgagcggaggcataaatcacgaaggaatcaaatattacaacaaccttatcaacgaactattggctaacggtatacaaccatttgtaactctttttcattgggatcttccccaagtcttagaagatgagtatggtggtttcttaaactccggtgtaataaatgattttcgagactatacggatctttgcttcaaggaatttggagatagagtgaggtattggagtactctaaatgagccatgggtgtttagcaattctggatatgcactaggaacaaatgcaccaggtcgatgttcggcctccaacgtggccaagcctggtgattctggaacaggaccttatatagttacacacaatcaaattcttgctcatgcagaagctgtacatgtgtataagactaaataccaggcatatcaaaagggaaagataggcataacgttggtatctaactggttaatgccacttgatgataatagcataccagatataaaggctgccgagagatcacttgacttccaatttggattgtttatggaacaattaacaacaggagattattctaagagcatgcggcgtatagttaaaaaccgattacctaagttctcaaaattcgaatcaagcctagtgaatggttcatttgattttattggtataaactattactcttctagttatattagcaatgccccttcacatggcaatgccaaacccagttactcaacaaatcctatgaccaatatttcatttgaaaaacatgggatacccttaggtccaagggctgcttcaatttggatatatgtttatccatatatgtttatccaagaggacttcgagatcttttgttacatattaaaaataaatataacaatcctgcaattttcaatcactgaaaatggtatgaatgaattcaacgatgcaacacttccagtagaagaagctcttttgaatacttacagaattgattactattaccgtcacttatactacattcgttctgcaatcagggctggctcaaatgtgaagggtttttacgcatggtcatttttggactgtaatgaatggtttgcaggctttactgttcgttttggattaaactttgtagattagaaagatggattaaaaaggtaccctaagctttctgcccaatggtacaagaactttctcaaaagaaactagctagtattattaaaagaactttgtagtagattacagtacatcgtttgaagttgagttggtgcacctaattaaataaaagaggttactcttaacatatttttaggccattcgttgtgaagttgttaggctgttatttctattatactatgttgtagtaataagtgcattgttgtaccagaagctatgatcataactataggttgatccttcatgtatcagtttgatgttgagaatactttgaattaaaagtctttttttatttttttaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
    'aaacaaaccaaatatggattttattgtagccatatttgctctgtttgttattagctcattcacaattacttccacaaatgcagttgaagcttctactcttcttgacataggtaacctgagtcggagcagttttcctcgtggcttcatctttggtgctggatcttcagcataccaatttgaaggtgcagtaaacgaaggcggtagaggaccaagtatttgggataccttcacccataaatatccagaaaaaataagggatggaagcaatgcagacatcacggttgaccaatatcaccgctacaaggaagatgttgggattatgaaggatcaaaatatggattcgtatagattctcaatctcttggccaagaatactcccaaagggaaagttgagcggaggcataaatcacgaaggaatcaaatattacaacaaccttatcaacgaactattggctaacggtatacaaccatttgtaactctttttcattgggatcttccccaagtcttagaagatgagtatggtggtttcttaaactccggtgtaataaatgattttcgagactatacggatctttgcttcaaggaatttggagatagagtgaggtattggagtactctaaatgagccatgggtgtttagcaattctggatatgcactaggaacaaatgcaccaggtcgatgttcggcctccaacgtggccaagcctggtgattctggaacaggaccttatatagttacacacaatcaaattcttgctcatgcagaagctgtacatgtgtataagactaaataccaggcatatcaaaagggaaagataggcataacgttggtatctaactggttaatgccacttgatgataatagcataccagatataaaggctgccgagagatcacttgacttccaatttggattgtttatggaacaattaacaacaggagattattctaagagcatgcggcgtatagttaaaaaccgattacctaagttctcaaaattcgaatcaagcctagtgaatggttcatttgattttattggtataaactattactcttctagttatattagcaatgccccttcacatggcaatgccaaacccagttactcaacaaatcctatgaccaatatttcatttgaaaaacatgggatacccttaggtccaagggctgcttcaatttggatatatgtttatccatatatgtttatccaagaggacttcgagatcttttgttacatattaaaaataaatataacaatcctgcaattttcaatcactgaaaatggtatgaatgaattcaacgatgcaacacttccagtagaagaagctcttttgaatacttacagaattgattactattaccgtcacttatactacattcgttctgcaatcagggctggctcaaatgtgaagggtttttacgcatggtcatttttggactgtaatgaatggtttgcaggctttactgttcgttttggattaaactttgtagattagaaagatggattaaaaaggtaccctaagctttctgcccaatggtacaagaactttctcaaaagaaactagctagtattattaaaagaactttgtagtagattacagtacatcgtttgaagttgagttggtgcacctaattaaataaaagaggttactcttaacatatttttaggccattcgttgtgaagttgttaggctgttatttctattatactatgttgtagtaataagtgcattgttgtaccagaagctatgatcataactataggttgatccttcatgtatcagtttgatgttgagaatactttgaattaaaagtctttttttatttttttaaaaaaaaaaaaaaaaaaaaccccccccc',
]
class TestFasta(unittest.TestCase):
    def setUp(self):
        self.fasta = sequences.Fasta('ID', 'ACGTA')

    def test_equality(self):
        self.assertTrue(self.fasta == sequences.Fasta('ID', 'ACGTA'))
        self.assertFalse(self.fasta == sequences.Fasta('I', 'ACGTA'))
        self.assertFalse(self.fasta == sequences.Fasta('ID', 'ACGT'))
        self.assertFalse(self.fasta != sequences.Fasta('ID', 'ACGTA'))
        self.assertTrue(self.fasta != sequences.Fasta('I', 'ACGTA'))
        self.assertTrue(self.fasta != sequences.Fasta('ID', 'ACGT'))

    def test_init(self):
        '''__init__ should get the ID and sequence correctly'''
        self.assertEqual(self.fasta.id, 'ID')
        self.assertEqual(self.fasta.seq, 'ACGTA')

    def test_get_next_from_file(self):
        '''get_next_from_file() should read seqs from OK, including weirdness in file'''
        f_in = utils.open_file_read(os.path.join(data_dir, 'sequences_test.fa'))
        fa = sequences.Fasta()
        counter = 1

        while fa.get_next_from_file(f_in):
            self.assertEqual(fa, sequences.Fasta(str(counter), 'ACGTA'))
            counter += 1

        utils.close(f_in)

    def test_get_id_from_header_line(self):
        '''Check that can get ID from header line or die properly'''
        self.assertEqual(sequences.Fasta._get_id_from_header_line(self.fasta, '>X'), 'X')
        with self.assertRaises(sequences.Error):
            self.assertEqual(sequences.Fasta._get_id_from_header_line(self.fasta, 'X'), 'X')

    def test_getitem(self):
        '''getitem() should return the right subsequence'''
        seq = 'AACGTGTCA'
        fa = sequences.Fasta('x', seq)
        self.assertEqual(seq[1], fa[1])
        self.assertEqual(seq[0:2], fa[0:2])
        self.assertEqual(seq[1:], fa[1:])

    def test_len(self):
        '''len() should return the length of the sequence'''
        self.assertEqual(5, len(self.fasta))

    def test_subseq(self):
        '''Test subseq'''
        fa = sequences.Fasta('name', 'ACGTA')
        self.assertEqual(fa.subseq(1,4), sequences.Fasta('name', 'CGT'))
        self.assertEqual(fa.subseq(None,4), sequences.Fasta('name', 'ACGT'))
        self.assertEqual(fa.subseq(1,None), sequences.Fasta('name', 'CGTA'))

    def test_print_line_length(self):
        '''__str__ should be formatted correctly with the right number of chars per line of sequence'''
        line_lengths = [0, 3]
        correct_files = [os.path.join(data_dir, x) for x in ['sequences_test_one-per-line.fa', 'sequences_test_3-per-line.fa']]

        for i in range(len(line_lengths)):
            seq_reader = sequences.file_reader(os.path.join(data_dir, 'sequences_test_one-per-line.fa'))
            sequences.Fasta.line_length = line_lengths[i]
            tmp_out = 'tmp.line_length_test.fa'
            f = utils.open_file_write(tmp_out)
            for s in seq_reader:
                print(s, file=f)
            utils.close(f)
            self.assertTrue(filecmp.cmp(correct_files[i], tmp_out))
            os.unlink(tmp_out)

        sequences.Fasta.line_length = 60

    def test_strip_after_first_whitespace(self):
        '''Test strip_after_first_whitespace()'''
        seqs = [
            sequences.Fasta('name', 'A'),
            sequences.Fasta('name foo', 'A'),
            sequences.Fasta('name foo bar', 'A'),
            sequences.Fasta('name\tfoo', 'A'),
        ]

        for seq in seqs:
            seq.strip_after_first_whitespace()

        for seq in seqs:
            self.assertEqual(seq.id, 'name')

    def test_strip_illumina_suffix(self):
        '''Check that /1 and /2 removed correctly from IDs'''
        seqs = [sequences.Fasta('name/1', 'A'),
                sequences.Fasta('name/2', 'A'),
                sequences.Fasta('name', 'A'),
                sequences.Fasta('name/1/2', 'A'),
                sequences.Fasta('name/2/1', 'A'),
                sequences.Fasta('name/3', 'A')]

        correct_names = ['name', 'name', 'name', 'name/1', 'name/2', 'name/3']

        for seq in seqs:
            seq.strip_illumina_suffix()

        for i in range(len(seqs)):
            self.assertEqual(seqs[i].id, correct_names[i])

    def test_revcomp(self):
        '''revcomp() should correctly reverse complement a sequence'''
        fa = sequences.Fasta('ID', 'ACGTNacgtn')
        fa.revcomp()
        self.assertEqual(fa, sequences.Fasta('ID', 'nacgtNACGT'))

    def test_gaps(self):
        '''gaps() should find the gaps in a sequence correctly'''
        test_seqs = [sequences.Fasta('ID', 'ACGT'),
                     sequences.Fasta('ID', 'NACGT'),
                     sequences.Fasta('ID', 'NACGTN'),
                     sequences.Fasta('ID', 'ANNCGT'),
                     sequences.Fasta('ID', 'NANNCGTNN')]

        correct_gaps = [[],
                        [intervals.Interval(0, 0)],
                        [intervals.Interval(0, 0), intervals.Interval(5, 5)],
                        [intervals.Interval(1, 2)],
                        [intervals.Interval(0, 0), intervals.Interval(2, 3), intervals.Interval(7, 8)]]

        for i in range(len(test_seqs)):
            gaps = test_seqs[i].gaps()
            self.assertListEqual(correct_gaps[i], gaps)

    def test_contig_coords(self):
        '''contig_coords() should get the coords of all contigs in a sequence correctly'''
        test_seqs = [sequences.Fasta('ID', 'ACGT'),
                     sequences.Fasta('ID', 'NACGT'),
                     sequences.Fasta('ID', 'NNACGT'),
                     sequences.Fasta('ID', 'ACGTN'),
                     sequences.Fasta('ID', 'ACGTNN'),
                     sequences.Fasta('ID', 'NANNCGT'),
                     sequences.Fasta('ID', 'ACNNNGTNA'),
                     sequences.Fasta('ID', 'ANNCGTNNAAAAA')]

        correct_coords = [[intervals.Interval(0,3)],
                         [intervals.Interval(1, 4)],
                         [intervals.Interval(2, 5)],
                         [intervals.Interval(0, 3)],
                         [intervals.Interval(0, 3)],
                         [intervals.Interval(1, 1), intervals.Interval(4,6)],
                         [intervals.Interval(0, 1), intervals.Interval(5, 6), intervals.Interval(8, 8)],
                         [intervals.Interval(0, 0), intervals.Interval(3, 5), intervals.Interval(8, 12)]]

        for i in range(len(test_seqs)):
            gaps = test_seqs[i].contig_coords()
            self.assertListEqual(correct_coords[i], gaps)




    def test_orfs(self):
        '''Test orfs()'''
        test_seqs = [(sequences.Fasta('ID', 'AAACCCGG'), 0, False, [intervals.Interval(0,5)]),
                     (sequences.Fasta('ID', 'AAAACCCGG'), 1, False, [intervals.Interval(1,6)]),
                     (sequences.Fasta('ID', 'AAAAACCCGG'), 2, False, [intervals.Interval(2,7)]),
                     (sequences.Fasta('ID', 'CCGGGTTT'), 0, True, [intervals.Interval(2,7)]),
                     (sequences.Fasta('ID', 'CCGGGTTTT'), 1, True, [intervals.Interval(2,7)]),
                     (sequences.Fasta('ID', 'CCGGGTTTTT'), 2, True, [intervals.Interval(2,7)]),
                     (sequences.Fasta('ID', 'AAACCCTGA'), 0, False, [intervals.Interval(0,8)]),
                     (sequences.Fasta('ID', 'AAACCCTGATAG'), 0, False, [intervals.Interval(0,8)]),
                     (sequences.Fasta('ID', 'AAACCCTGA'), 1, False, [intervals.Interval(1,6)]),
                     (sequences.Fasta('ID', ''), 0, False, []),
                     (sequences.Fasta('ID', 'A'), 0, False, []),
                     (sequences.Fasta('ID', 'AA'), 0, False, []),
                     (sequences.Fasta('ID', 'AAA'), 0, False, [intervals.Interval(0,2)]),
                     (sequences.Fasta('ID', 'AAAAAA'), 0, False, [intervals.Interval(0,5)]),
                     (sequences.Fasta('ID', 'AAA'), 1, False, []),
                     (sequences.Fasta('ID', 'AAA'), 2, False, []),
                     (sequences.Fasta('ID', 'AAA'), 0, True, [intervals.Interval(0,2)]),
                     (sequences.Fasta('ID', 'AAA'), 1, True, []),
                     (sequences.Fasta('ID', 'AAA'), 2, True, []),
                     (sequences.Fasta('ID', 'TAA'), 0, False, []),
                     (sequences.Fasta('ID', 'CTA'), 0, True, [])]


        for t in test_seqs:
            orfs = t[0].orfs(frame=t[1], revcomp=t[2])
            self.assertListEqual(orfs, t[3])

    def test_all_orfs(self):
        '''Test all_orfs()'''
        d = {}
        tasks.file_to_dict(os.path.join(data_dir, 'sequences_test_orfs.fa'), d)
        seq = d['1']
        orfs = seq.all_orfs(min_length=120)
        expected = [
            (intervals.Interval(27, 221), False),
            (intervals.Interval(44, 226), False),
            (intervals.Interval(48, 170), True),
            (intervals.Interval(109, 240), False),
            (intervals.Interval(143, 265), True),
            (intervals.Interval(227, 421), False),
            (intervals.Interval(277, 432), True),
            (intervals.Interval(286, 477), False),
            (intervals.Interval(288, 518), True),
            (intervals.Interval(562, 702), False),
            (intervals.Interval(600, 758), False),
            (intervals.Interval(605, 817), False),
            (intervals.Interval(818, 937), False),
            (intervals.Interval(835, 987), False),
            (intervals.Interval(864, 998), False)
        ]

        self.assertEqual(len(orfs), len(expected))

        for i in range(len(orfs)):
            print(orfs[i][0], expected[i][0])
            self.assertEqual(orfs[i][0], expected[i][0])
            self.assertEqual(orfs[i][1], expected[i][1])


    def test_is_complete_orf(self):
        '''Test is_complete_orf'''
        tests = [
            (sequences.Fasta('ID', 'TTT'), False),
            (sequences.Fasta('ID', 'TTTTAA'), True),
            (sequences.Fasta('ID', 'TTTTAATAA'), False),
            (sequences.Fasta('ID', 'TTGTAA'), True),
            (sequences.Fasta('ID', 'TTTAAC'), True),
            (sequences.Fasta('ID', 'TGA'), False),
            (sequences.Fasta('ID', 'TGAA'), False),
        ]

        for t in tests:
            self.assertEqual(t[0].is_complete_orf(), t[1])


    def test_looks_like_gene(self):
        '''Test looks_like_gene'''
        tests = [
            (sequences.Fasta('ID', 'TTT'), False),
            (sequences.Fasta('ID', 'TTGTAA'), True),
            (sequences.Fasta('ID', 'ttgTAA'), True),
            (sequences.Fasta('ID', 'TTGTTTTAA'), True),
            (sequences.Fasta('ID', 'TTGTAATTTTAA'), False),
            (sequences.Fasta('ID', 'TTGTTTTGAA'), False),
        ]

        for t in tests:
            self.assertEqual(t[0].looks_like_gene(), t[1])

        sequences.genetic_code = 1
        self.assertFalse(sequences.Fasta('ID', 'ATTCAGTAA').looks_like_gene())
        sequences.genetic_code = 11
        self.assertTrue(sequences.Fasta('ID', 'ATTCAGTAA').looks_like_gene())
        sequences.genetic_code = 1


    def test_make_into_gene_fasta(self):
        '''Test make_into_gene fasta'''
        print('sequences.genetic_code', sequences.genetic_code)
        tests = [
            (sequences.Fasta('ID', 'T'), None),
            (sequences.Fasta('ID', 'TT'), None),
            (sequences.Fasta('ID', 'TTT'), None),
            (sequences.Fasta('ID', 'TTG'), None),
            (sequences.Fasta('ID', 'TAA'), None),
            (sequences.Fasta('ID', 'TTGAAATAA'), (sequences.Fasta('ID', 'TTGAAATAA'), '+', 0)),
            (sequences.Fasta('ID', 'TTGAAATAT'), None),
            (sequences.Fasta('ID', 'TTGTAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 0)),
            (sequences.Fasta('ID', 'TTGTAAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 0)),
            (sequences.Fasta('ID', 'TTGTAAAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 0)),
            (sequences.Fasta('ID', 'TTGTAAAAA'), None),
            (sequences.Fasta('ID', 'ATTGTAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 1)),
            (sequences.Fasta('ID', 'ATTGTAAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 1)),
            (sequences.Fasta('ID', 'ATTGTAAAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 1)),
            (sequences.Fasta('ID', 'ATTGTAAAAA'), None),
            (sequences.Fasta('ID', 'AATTGTAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 2)),
            (sequences.Fasta('ID', 'AATTGTAAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 2)),
            (sequences.Fasta('ID', 'AATTGTAAAA'), (sequences.Fasta('ID', 'TTGTAA'), '+', 2)),
            (sequences.Fasta('ID', 'AATTGTAAAAA'), None),
            (sequences.Fasta('ID', 'TTACAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 0)),
            (sequences.Fasta('ID', 'ATTACAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 0)),
            (sequences.Fasta('ID', 'AATTACAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 0)),
            (sequences.Fasta('ID', 'AAATTACAA'), None),
            (sequences.Fasta('ID', 'TTACAAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 1)),
            (sequences.Fasta('ID', 'ATTACAAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 1)),
            (sequences.Fasta('ID', 'AATTACAAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 1)),
            (sequences.Fasta('ID', 'AAATTACAAA'), None),
            (sequences.Fasta('ID', 'TTACAAAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 2)),
            (sequences.Fasta('ID', 'ATTACAAAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 2)),
            (sequences.Fasta('ID', 'AATTACAAAA'), (sequences.Fasta('ID', 'TTGTAA'), '-', 2)),
            (sequences.Fasta('ID', 'AAATTACAAAA'), None),
        ]

        for seq, expected in tests:
            self.assertEqual(seq.make_into_gene(), expected)


    def test_make_into_gene_fastq(self):
        '''Test make_into_gene fastq'''
        print('sequences.genetic_code', sequences.genetic_code)
        tests = [
            (sequences.Fastq('ID', 'T', '1'), None),
            (sequences.Fastq('ID', 'TT', '12'), None),
            (sequences.Fastq('ID', 'TTT', '123'), None),
            (sequences.Fastq('ID', 'TTG', '123'), None),
            (sequences.Fastq('ID', 'TAA', '123'), None),
            (sequences.Fastq('ID', 'TTGAAATAA', '123456789'), (sequences.Fastq('ID', 'TTGAAATAA', '123456789'), '+', 0)),
            (sequences.Fastq('ID', 'TTGAAATAT', '123456789'), None),
            (sequences.Fastq('ID', 'TTGTAA', '123456'), (sequences.Fastq('ID', 'TTGTAA', '123456'), '+', 0)),
            (sequences.Fastq('ID', 'TTGTAAA', '1234567'), (sequences.Fastq('ID', 'TTGTAA', '123456'), '+', 0)),
            (sequences.Fastq('ID', 'TTGTAAAA', '12345678'), (sequences.Fastq('ID', 'TTGTAA', '123456'), '+', 0)),
            (sequences.Fastq('ID', 'TTGTAAAAA', '123456789'), None),
            (sequences.Fastq('ID', 'ATTGTAA', '1234567'), (sequences.Fastq('ID', 'TTGTAA', '234567'), '+', 1)),
            (sequences.Fastq('ID', 'ATTGTAAA', '12345678'), (sequences.Fastq('ID', 'TTGTAA', '234567'), '+', 1)),
            (sequences.Fastq('ID', 'ATTGTAAAA', '123456789'), (sequences.Fastq('ID', 'TTGTAA', '234567'), '+', 1)),
            (sequences.Fastq('ID', 'ATTGTAAAAA', '123456789A'), None),
            (sequences.Fastq('ID', 'AATTGTAA', '12345678'), (sequences.Fastq('ID', 'TTGTAA', '345678'), '+', 2)),
            (sequences.Fastq('ID', 'AATTGTAAA', '123456789'), (sequences.Fastq('ID', 'TTGTAA', '345678'), '+', 2)),
            (sequences.Fastq('ID', 'AATTGTAAAA', '123456789A'), (sequences.Fastq('ID', 'TTGTAA', '345678'), '+', 2)),
            (sequences.Fastq('ID', 'AATTGTAAAAA', '123456789AB'), None),
            (sequences.Fastq('ID', 'TTACAA', '123456'), (sequences.Fastq('ID', 'TTGTAA', '654321'), '-', 0)),
            (sequences.Fastq('ID', 'ATTACAA', '1234567'), (sequences.Fastq('ID', 'TTGTAA', '765432'), '-', 0)),
            (sequences.Fastq('ID', 'AATTACAA', '12345678'), (sequences.Fastq('ID', 'TTGTAA', '876543'), '-', 0)),
            (sequences.Fastq('ID', 'AAATTACAA', '123456789'), None),
            (sequences.Fastq('ID', 'TTACAAA', '1234567'), (sequences.Fastq('ID', 'TTGTAA', '654321'), '-', 1)),
            (sequences.Fastq('ID', 'ATTACAAA', '12345678'), (sequences.Fastq('ID', 'TTGTAA', '765432'), '-', 1)),
            (sequences.Fastq('ID', 'AATTACAAA', '123456789'), (sequences.Fastq('ID', 'TTGTAA', '876543'), '-', 1)),
            (sequences.Fastq('ID', 'AAATTACAAA', '123456789A'), None),
            (sequences.Fastq('ID', 'TTACAAAA', '12345678'), (sequences.Fastq('ID', 'TTGTAA', '654321'), '-', 2)),
            (sequences.Fastq('ID', 'ATTACAAAA', '123456789'), (sequences.Fastq('ID', 'TTGTAA', '765432'), '-', 2)),
            (sequences.Fastq('ID', 'AATTACAAAA', '123456789A'), (sequences.Fastq('ID', 'TTGTAA', '876543'), '-', 2)),
            (sequences.Fastq('ID', 'AAATTACAAAA', '123456789AB'), None),
        ]

        for seq, expected in tests:
            self.assertEqual(seq.make_into_gene(), expected)

    def test_is_all_Ns(self):
        '''Test is_all_Ns()'''
        self.assertTrue(sequences.Fasta('ID', 'n').is_all_Ns())
        self.assertTrue(sequences.Fasta('ID', 'N').is_all_Ns())
        self.assertTrue(sequences.Fasta('ID', 'nNn').is_all_Ns())
        self.assertFalse(sequences.Fasta('ID', 'a').is_all_Ns())
        self.assertFalse(sequences.Fasta('ID', '').is_all_Ns())
        self.assertFalse(sequences.Fasta('ID', 'anNg').is_all_Ns())
        self.assertFalse(sequences.Fasta('ID', 'naN').is_all_Ns())
        self.assertFalse(sequences.Fasta('ID', 'anNg').is_all_Ns(start=0, end=0))
        self.assertFalse(sequences.Fasta('ID', 'anNg').is_all_Ns(start=0, end=1))
        self.assertTrue(sequences.Fasta('ID', 'anNg').is_all_Ns(start=1, end=1))
        self.assertTrue(sequences.Fasta('ID', 'anNg').is_all_Ns(start=1, end=2))
        self.assertFalse(sequences.Fasta('ID', 'anNg').is_all_Ns(start=1))
        self.assertTrue(sequences.Fasta('ID', 'anN').is_all_Ns(start=1))
        self.assertFalse(sequences.Fasta('ID', 'anNg').is_all_Ns(end=1))
        self.assertTrue(sequences.Fasta('ID', 'nNA').is_all_Ns(end=1))

        with self.assertRaises(sequences.Error):
            sequences.Fasta('ID', 'anNg').is_all_Ns(start=1, end=0)

    def test_trim_Ns(self):
        '''trim_Ns() should do the right trimming of a sequence'''
        fa = sequences.Fasta('ID', 'ANNANA')
        test_seqs = [sequences.Fasta('ID', 'ANNANA'),
                     sequences.Fasta('ID', 'NANNANA'),
                     sequences.Fasta('ID', 'NANNANAN'),
                     sequences.Fasta('ID', 'ANNANAN'),
                     sequences.Fasta('ID', 'NNNNNNANNANAN'),
                     sequences.Fasta('ID', 'NNANNANANn')]

        for s in test_seqs:
            s.trim_Ns()
            self.assertEqual(fa, s)

    def test_add_insertions(self):
        '''Test add_insertions'''
        fa = sequences.Fasta('X', 'acgtacgtacgt')
        fa.add_insertions(skip=4, window=0, test=True)
        self.assertEqual(fa, sequences.Fasta('X', 'acgtNacgtNacgt'))

    def test_replace_bases(self):
        '''Check that bases get replaced correctly'''
        fa = sequences.Fasta('X', 'AUCGTUUACT')
        fa.replace_bases('U', 'T')
        self.assertEqual(fa, sequences.Fasta('X', 'ATCGTTTACT'))


    def test_replace_non_acgt(self):
        '''test replace_non_acgt'''
        tests = [
            ('acgtACGTnN', 'acgtACGTnN'),
            ('abc.g-T?aRC1T', 'aNcNgNTNaNCNT')
        ]

        for seq, expected in tests:
            fa = sequences.Fasta('id', seq)
            fa.replace_non_acgt()
            self.assertEqual(expected, fa.seq)


    def test_replace_interval(self):
        '''Test replace_interval()'''
        fa = sequences.Fasta('ID', 'ACGTA')
        fa.replace_interval(0, 0, 'NEW')
        self.assertEqual(fa, sequences.Fasta('ID', 'NEWCGTA'))

        fa = sequences.Fasta('ID', 'ACGTA')
        fa.replace_interval(4, 4, 'NEW')
        self.assertEqual(fa, sequences.Fasta('ID', 'ACGTNEW'))

        fa = sequences.Fasta('ID', 'ACGTA')
        fa.replace_interval(2, 3, 'NEW')
        self.assertEqual(fa, sequences.Fasta('ID', 'ACNEWA'))

        fa = sequences.Fasta('ID', 'ACGTA')
        with self.assertRaises(sequences.Error):
            fa.replace_interval(3,2,'x')
        with self.assertRaises(sequences.Error):
            fa.replace_interval(1,5,'x')
        with self.assertRaises(sequences.Error):
            fa.replace_interval(5,10,'x')

        fq = sequences.Fastq('ID', 'ACGTA', 'ABCDE')
        fq.replace_interval(0, 0, 'NEW', 'III')
        self.assertEqual(fq, sequences.Fastq('ID', 'NEWCGTA', 'IIIBCDE'))

        fq = sequences.Fastq('ID', 'ACGTA', 'ABCDE')
        fq.replace_interval(4, 4, 'NEW', 'III')
        self.assertEqual(fq, sequences.Fastq('ID', 'ACGTNEW', 'ABCDIII'))

        fq = sequences.Fastq('ID', 'ACGTA', 'ABCDE')
        fq.replace_interval(2, 3, 'NEW', 'III')
        self.assertEqual(fq, sequences.Fastq('ID', 'ACNEWA', 'ABIIIE'))

        with self.assertRaises(sequences.Error):
            fq.replace_interval(1,1,'x', 'xx')

    def test_search_string(self):
        '''Check that search_string() finds all the hits'''
        fa = sequences.Fasta('X', 'AAA')
        hits = fa.search('G')
        self.assertTrue(len(hits) == 0)
        hits = fa.search('AAA')
        self.assertListEqual(hits, [(0, '+')])
        hits = fa.search('AA')
        self.assertListEqual(hits, [(0, '+'), (1, '+')])
        hits = fa.search('TTT')
        self.assertListEqual(hits, [(0, '-')])

    def test_to_Fastq(self):
        '''Check to_Fastq converts OK, including out of range quality scores'''
        fa = sequences.Fasta('X', 'AAAAA')
        quals = [-1, 0, 40, 93, 94]
        self.assertEqual(sequences.Fastq('X', 'AAAAA', '!!I~~'), fa.to_Fastq(quals))
        with self.assertRaises(sequences.Error):
            fa.to_Fastq('AAAAAAAAAAAAA')


    def test_translate(self):
        '''Test nucleotide -> amino acid conversion works on Fasta'''
        fa = sequences.Fasta('ID', 'GCAGCCGCGGCTAGAAGGCGACGCCGGCGTAACAATGACGATTGCTGTGAAGAGCAACAGGGAGGCGGGGGTCACCATATAATCATTTTATTGCTACTCCTGCTTAAAAAGATGTTCTTTCCACCCCCGCCTAGCAGTTCATCCTCGTCTACAACCACGACTTGGTACTATGTAGTCGTGGTTTAATAGTGA')
        self.assertEqual(sequences.Fasta('ID', 'AAAARRRRRRNNDDCCEEQQGGGGHHIIILLLLLLKKMFFPPPPSSSSSSTTTTWYYVVVV***'), fa.translate())
        self.assertEqual(sequences.Fasta('ID', 'QPRLEGDAGVTMTIAVKSNREAGVTI*SFYCYSCLKRCSFHPRLAVHPRLQPRLGTM*SWFNS'), fa.translate(frame=1))
        print(fa.translate(frame=1))
        self.assertEqual(sequences.Fasta('ID', 'SRG*KATPA*Q*RLL*RATGRRGSPYNHFIATPA*KDVLSTPA*QFILVYNHDLVLCSRGLIV'), fa.translate(frame=2))

    def test_expand_nucleotides(self):
        '''Test expand_nucleotides'''
        tests = [
            (sequences.Fasta('1', 'A'), [sequences.Fasta('1.1', 'A')]),
            (sequences.Fasta('2', 'C'), [sequences.Fasta('2.1', 'C')]),
            (sequences.Fasta('3', 'G'), [sequences.Fasta('3.1', 'G')]),
            (sequences.Fasta('4', 'T'), [sequences.Fasta('4.1', 'T')]),
            (sequences.Fasta('6', 'R'), [sequences.Fasta('6.1', 'A'), sequences.Fasta('6.2', 'G')]),
            (sequences.Fasta('7', 'Y'), [sequences.Fasta('7.1', 'C'), sequences.Fasta('7.2', 'T')]),
            (sequences.Fasta('8', 'S'), [sequences.Fasta('8.1', 'C'), sequences.Fasta('8.2', 'G')]),
            (sequences.Fasta('9', 'W'), [sequences.Fasta('9.1', 'A'), sequences.Fasta('9.2', 'T')]),
            (sequences.Fasta('10', 'K'), [sequences.Fasta('10.1', 'G'), sequences.Fasta('10.2', 'T')]),
            (sequences.Fasta('11', 'M'), [sequences.Fasta('11.1', 'A'), sequences.Fasta('11.2', 'C')]),
            (sequences.Fasta('12', 'B'), [sequences.Fasta('12.1', 'C'), sequences.Fasta('12.2', 'G'), sequences.Fasta('12.3', 'T')]),
            (sequences.Fasta('13', 'D'), [sequences.Fasta('13.1', 'A'), sequences.Fasta('13.2', 'G'), sequences.Fasta('13.3', 'T')]),
            (sequences.Fasta('14', 'H'), [sequences.Fasta('14.1', 'A'), sequences.Fasta('14.2', 'C'), sequences.Fasta('14.3', 'T')]),
            (sequences.Fasta('15', 'V'), [sequences.Fasta('15.1', 'A'), sequences.Fasta('15.2', 'C'), sequences.Fasta('15.3', 'G')]),
            (sequences.Fasta('16', 'N'), [sequences.Fasta('16.1', 'A'), sequences.Fasta('16.2', 'C'), sequences.Fasta('16.3', 'G'), sequences.Fasta('16.4', 'T')]),
            (sequences.Fasta('17', 'ART'), [sequences.Fasta('17.1', 'AAT'), sequences.Fasta('17.2', 'AGT')]),
            (sequences.Fasta('18', 'ARRT'), [sequences.Fasta('18.1', 'AAAT'), sequences.Fasta('18.2', 'AAGT'), sequences.Fasta('18.3', 'AGAT'), sequences.Fasta('18.4', 'AGGT')]),
            (sequences.Fasta('19', 'ARTR'), [sequences.Fasta('19.1', 'AATA'), sequences.Fasta('19.2', 'AATG'), sequences.Fasta('19.3', 'AGTA'), sequences.Fasta('19.4', 'AGTG')]),
            (sequences.Fastq('20', 'ART', 'GHI'), [sequences.Fastq('20.1', 'AAT', 'GHI'), sequences.Fastq('20.2', 'AGT', 'GHI')]),
        ]

        for t in tests:
            self.assertListEqual(t[0].expand_nucleotides(), t[1])

    def test_split_capillary_id(self):
        '''Tests that we get information from a sanger capillary read name OK'''
        ids = ['abcde.p1k', 'abcde.x.p1k', 'abcde.p1ka', 'abcde.q1k', 'abcde.w2k']
        expected = [{'prefix': 'abcde', 'dir': 'fwd', 'suffix': 'p1k'},
                    {'prefix': 'abcde.x', 'dir': 'fwd', 'suffix': 'p1k'},
                    {'prefix': 'abcde', 'dir': 'fwd', 'suffix': 'p1ka'},
                    {'prefix': 'abcde', 'dir': 'rev', 'suffix': 'q1k'},
                    {'prefix': 'abcde', 'dir': 'unk', 'suffix': 'w2k'}]

        for i in range(len(ids)):
            fa = sequences.Fasta(ids[i], 'A')
            self.assertEqual(fa.split_capillary_id(), expected[i])

        with self.assertRaises(sequences.Error):
            fa = sequences.Fasta('name', 'A')
            fa.split_capillary_id()


class TestEmbl(unittest.TestCase):
    def test_get_id_from_header_line(self):
        '''Test get id from header line of EMBL'''
        embl = sequences.Embl('ID', 'ACGT')
        self.assertEqual(embl._get_id_from_header_line('ID   X; blah'), 'X')
        self.assertEqual(embl._get_id_from_header_line('LOCUS   X foo'), 'X')
        with self.assertRaises(sequences.Error):
            self.assertEqual(embl._get_id_from_header_line('ID X;'), 'X')
        with self.assertRaises(sequences.Error):
            self.assertEqual(embl._get_id_from_header_line('XX   X;'), 'X')


    def test_get_next_from_embl_file(self):
        f_in = utils.open_file_read(os.path.join(data_dir, 'sequences_test.embl'))
        embl = sequences.Embl()
        counter = 1

        while embl.get_next_from_file(f_in):
            self.assertEqual(embl, sequences.Fasta('seq' + str(counter), expected_embl[counter-1]))
            counter += 1

        utils.close(f_in)


    def test_get_next_from_gbk_file(self):
        f_in = utils.open_file_read(os.path.join(data_dir, 'sequences_test.gbk'))
        embl = sequences.Embl()
        counter = 1
        expected = [
            'gatcctccatatacaacggtatctccacctcaggtttagatctcaacaacggaaccattgccgacatgagacagttaggtatcgtcgagagttacaagctaaaacgagcagtagtcagctctgcatctgaagccgctgaagttctactaagggtggataacatcatccgtgcaagaccaatgccatgactcagattctaattttaagctattcaatttctctttgatc',
            'gatcctccatatacaacggtatctccacctcaggtttagatctcaacaacggaaccattgccgacatgagacagttaggtatcgtcgagagttacaagctaaaacgagcagtagtcagctctgcatctgaagccgctgaagttctactaagggtggataacatcatccgtgcaagaccaatgccatgactcagattctaattttaagctattcaatttctctttgaaa']

        while embl.get_next_from_file(f_in):
            self.assertEqual(embl, sequences.Fasta('NAME' + str(counter), expected[counter-1]))
            counter += 1

        utils.close(f_in)


class TestFastq(unittest.TestCase):
    def setUp(self):
        self.fastq = sequences.Fastq('ID', 'ACGTA', 'IIIII')

    def test_init(self):
        '''__init__ should get the ID, sequence and quality correctly'''
        self.assertEqual(self.fastq.id, 'ID')
        self.assertEqual(self.fastq.seq, 'ACGTA')
        self.assertEqual(self.fastq.qual, 'IIIII')

    def test_init_length_mismatch(self):
        '''__init__ should raise an error when length of seq and quality not the same'''
        with self.assertRaises(sequences.Error):
            sequences.Fastq('X', 'A', 'II')

    def test_get_next_from_file(self):
        '''get_next_from_file() should read seqs from OK, and raise error at badly formatted file'''
        bad_files = ['sequences_test_fail_no_AT.fq',
                     'sequences_test_fail_no_seq.fq',
                     'sequences_test_fail_no_plus.fq',
                     'sequences_test_fail_no_qual.fq']

        bad_files = [os.path.join(data_dir, x) for x in bad_files]

        for fname in bad_files:
            f_in = utils.open_file_read(fname)
            fq = sequences.Fastq()
            with self.assertRaises(sequences.Error):
                while fq.get_next_from_file(f_in):
                    pass

            utils.close(f_in)

        fname = os.path.join(data_dir, 'sequences_test_good_file.fq')
        try:
            f_in = open(fname)
        except IOError:
            print("Error opening '" + fname + "'", file=sys.stderr)
            sys.exit(1)

        fq = sequences.Fastq()
        while fq.get_next_from_file(f_in):
            self.assertEqual(fq, sequences.Fastq('ID', 'ACGTA', 'IIIII'))
        utils.close(f_in)

    def test_subseq(self):
        '''Test subseq'''
        fq = sequences.Fastq('name', 'ACGTA', 'FGHIJ')
        self.assertEqual(fq.subseq(1,4), sequences.Fastq('name', 'CGT', 'GHI'))
        self.assertEqual(fq.subseq(None,4), sequences.Fastq('name', 'ACGT', 'FGHI'))
        self.assertEqual(fq.subseq(1,None), sequences.Fastq('name', 'CGTA', 'GHIJ'))

    def test_revcomp(self):
        '''revcomp() should correctly reverse complement a sequence'''
        fq = sequences.Fastq('ID', 'ACGTNacgtn', '1234567890')
        fq.revcomp()
        self.assertEqual(fq, sequences.Fastq('ID', 'nacgtNACGT', '0987654321'))

    def test_trim_Ns(self):
        '''trim_Ns() should do the right trimming of a fastq sequence'''
        fq = sequences.Fastq('ID', 'ANNANA', '111111')
        test_seqs = [sequences.Fastq('ID', 'ANNANA', '111111'),
                     sequences.Fastq('ID', 'NANNANA', '1111111'),
                     sequences.Fastq('ID', 'NANNANAN', '11111111'),
                     sequences.Fastq('ID', 'ANNANAN', '1111111'),
                     sequences.Fastq('ID', 'NNNNNNANNANAN', '1111111111111'),
                     sequences.Fastq('ID', 'NNANNANANn', '1111111111')]

        for s in test_seqs:
            s.trim_Ns()
            self.assertEqual(fq, s)

    def test_trim(self):
        '''trim() should trim the right number of bases off start and end'''
        fq = sequences.Fastq('ID', '1234567890', '1234567890')
        fq.trim(0, 0)
        self.assertEqual(fq, sequences.Fastq('ID', '1234567890', '1234567890'))

        fq = sequences.Fastq('ID', '1234567890', '1234567890')
        fq.trim(1, 0)
        self.assertEqual(fq, sequences.Fastq('ID', '234567890', '234567890'))

        fq = sequences.Fastq('ID', '1234567890', '1234567890')
        fq.trim(0, 1)
        self.assertEqual(fq, sequences.Fastq('ID', '123456789', '123456789'))

        fq = sequences.Fastq('ID', '1234567890', '1234567890')
        fq.trim(2, 2)
        self.assertEqual(fq, sequences.Fastq('ID', '345678', '345678'))

    def test_to_Fasta_and_qual(self):
        '''Check to_Fasta_and_qual converts quality scores correctly'''
        fq = sequences.Fastq('ID', 'ACGT', '>ADI')
        (fa, qual) = fq.to_Fasta_and_qual()
        self.assertEqual(fa, sequences.Fasta('ID', 'ACGT'))
        self.assertListEqual(qual, [29, 32, 35, 40])


    def test_translate(self):
        '''Test nucleatide -> amino acid conversion works on Fasta'''
        fq = sequences.Fastq('ID', 'GCAGCCGCGGCTAGAAGGCGACGCCGGCGTAACAATGACGATTGCTGTGAAGAGCAACAGGGAGGCGGGGGTCACCATATAATCATTTTATTGCTACTCCTGCTTAAAAAGATGTTCTTTCCACCCCCGCCTAGCAGTTCATCCTCGTCTACAACCACGACTTGGTACTATGTAGTCGTGGTTTAATAGTGA', 'IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII')

        self.assertEqual(sequences.Fastq('ID', 'AAAARRRRRRNNDDCCEEQQGGGGHHIIILLLLLLKKMFFPPPPSSSSSSTTTTWYYVVVV***', 'IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII'), fq.translate())

class TestFileReader(unittest.TestCase):
    def test_file_reader_fasta(self):
        '''file_reader should iterate through a fasta file correctly'''
        reader = sequences.file_reader(os.path.join(data_dir, 'sequences_test.fa'))
        counter = 1
        for seq in reader:
            self.assertEqual(seq, sequences.Fasta(str(counter), 'ACGTA'))
            counter += 1

    def test_file_reader_fastq(self):
        '''file_reader should iterate through a fastq file correctly'''
        reader = sequences.file_reader(os.path.join(data_dir, 'sequences_test_good_file.fq'))
        for seq in reader:
            self.assertEqual(seq, sequences.Fastq('ID', 'ACGTA', 'IIIII'))

    def test_file_reader_bad_format(self):
        '''file_reader should die properly when not given fasta or fastq file'''
        with self.assertRaises(sequences.Error):
            reader = sequences.file_reader(os.path.join(data_dir, 'sequences_test_not_a_fastaq_file'))
            for seq in reader:
                pass

    def test_file_reader_gff(self):
        '''Test read gff file'''
        good_files = [
            'sequences_test_gffv3.gff',
            'sequences_test_gffv3.no_FASTA_line.gff'
        ]
        good_files = [os.path.join(data_dir, x) for x in good_files]

        for f in good_files:
            reader = sequences.file_reader(f)
            counter = 1
            for seq in reader:
                self.assertEqual(seq, sequences.Fasta('seq' + str(counter), 'ACGTACGTAC'))
                counter += 1

        bad_files = [
            'sequences_test_gffv3.no_seq.gff',
            'sequences_test_gffv3.no_seq.2.gff'
        ]
        bad_files = [os.path.join(data_dir, x) for x in bad_files]

        for filename in bad_files:
            with self.assertRaises(sequences.Error):
                reader = sequences.file_reader(filename)
                for seq in reader:
                    pass

    def test_file_reader_embl(self):
        '''Test read embl file'''
        reader = sequences.file_reader(os.path.join(data_dir, 'sequences_test.embl'))

        counter = 1
        for seq in reader:
            self.assertEqual(seq, sequences.Fasta('seq' + str(counter), expected_embl[counter-1]))
            counter += 1

        bad_files = [
            'sequences_test.embl.bad',
            'sequences_test.embl.bad2',
        ]
        bad_files = [os.path.join(data_dir, x) for x in bad_files]

        for filename in bad_files:
            with self.assertRaises(sequences.Error):
                reader = sequences.file_reader(filename)
                for seq in reader:
                    pass

    def test_file_reader_phylip(self):
        '''Test read phylip file'''
        test_files = [
            'sequences_test_phylip.interleaved',
            'sequences_test_phylip.interleaved2',
            'sequences_test_phylip.sequential'
        ]

        test_files = [os.path.join(data_dir, f) for f in test_files]

        expected_seqs = [
            sequences.Fasta('Turkey', 'AACTNGGGCATTTCAGGGTGAGCCCGGGCAATACAGGGTAT'),
            sequences.Fasta('Salmo_gair', 'AAGCCTTGGCAGTGCAGGGTGAGCCGTGGCCGGGCACGGTAT'),
            sequences.Fasta('H. Sapiens', 'ACCGGTTGGCCGTTCAGGGTACAGGTTGGCCGTTCAGGGTAA')
        ]

        for fname in test_files:
            reader = sequences.file_reader(fname)
            i = 0
            for seq in reader:
                self.assertEqual(expected_seqs[i], seq)
                i += 1

        # files made by seaview are a little different in the first line.
        # Test one of these
        expected_seqs = [
            sequences.Fasta('seq1', 96 * 'G' + 'T'),
            sequences.Fasta('seq2', 94 * 'A' + 'G')
        ]

        reader = sequences.file_reader(os.path.join(data_dir, 'sequences_test_phylip.made_by_seaview'))
        i = 0
        for seq in reader:
            print(seq)
            self.assertEqual(expected_seqs[i], seq)
            i += 1


class TestOther(unittest.TestCase):
    def test_orfs_from_aa_seq(self):
        '''Test _orfs_from_aa_seq()'''
        test_seqs = ['',
                     '*',
                     '**',
                     'A',
                     'A*A*A',
                     'AB**CDE*AB',
                     '*ABCDE*',
                     '**ABCDE**']

        correct_coords = [[],
                          [],
                          [],
                          [intervals.Interval(0, 0)],
                          [intervals.Interval(0, 1), intervals.Interval(2, 3),intervals.Interval(4, 4)],
                          [intervals.Interval(0, 2), intervals.Interval(4, 7), intervals.Interval(8, 9)],
                          [intervals.Interval(1, 6)],
                          [intervals.Interval(2, 7)]]

        for i in range(len(test_seqs)):
            orfs = sequences._orfs_from_aa_seq(test_seqs[i])
            self.assertListEqual(correct_coords[i], orfs)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import sys
import filecmp
import os
import unittest
from pyfastaq import tasks, sequences

modules_dir = os.path.dirname(os.path.abspath(sequences.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class Error (Exception): pass


class TestACGTN_only(unittest.TestCase):
    def test_acgtn_only(self):
        '''Test acgtn_only'''
        tmpfile = 'tmp.test_acgtn_only.fa'
        infile = os.path.join(data_dir, 'test_acgtn_only.in.fa')
        expected = os.path.join(data_dir, 'test_acgtn_only.expected.fa')
        tasks.acgtn_only(infile, tmpfile)
        self.assertTrue(filecmp.cmp(expected, tmpfile, shallow=False))
        os.unlink(tmpfile)


class TestCafToFastq(unittest.TestCase):
    def test_caf_to_fastq_default(self):
        '''Test caf_to_fastq with no filtering'''
        tmpfile = 'tmp.fq'
        tasks.caf_to_fastq(os.path.join(data_dir, 'caf_test.caf'), tmpfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'caf_test.to_fastq.no_trim.min_length_0.fq'), tmpfile, shallow=False))
        os.unlink(tmpfile)

    def test_caf_to_fastq_trim_and_min_length(self):
        '''Test caf_to_fastq with trimming and min_length'''
        tmpfile = 'tmp.fq'
        tasks.caf_to_fastq(os.path.join(data_dir, 'caf_test.caf'), tmpfile, trim=True, min_length=6)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'caf_test.to_fastq.trim.min_length_6.fq'), tmpfile, shallow=False))
        os.unlink(tmpfile)


class TestCapillaryToPairs(unittest.TestCase):
    def test_capillary_to_pairs(self):
        '''Check that capillary reads file converted to paired and unpaired'''
        tmp_prefix = 'tmp.cap_to_pairs'
        tasks.capillary_to_pairs(os.path.join(data_dir, 'sequences_test_cap_to_read_pairs.fa'), tmp_prefix)
        # sequences have been hashed, so could be in any order in
        # output files. So need to check contents of files are OK
        d_correct_paired = {}
        d_correct_unpaired = {}
        tasks.file_to_dict(os.path.join(data_dir, 'sequences_test_cap_to_read_pairs.fa.paired.gz'), d_correct_paired)
        tasks.file_to_dict(os.path.join(data_dir, 'sequences_test_cap_to_read_pairs.fa.unpaired.gz'), d_correct_unpaired)
        d_test_paired = {}
        d_test_unpaired = {}
        tasks.file_to_dict(tmp_prefix + '.paired.gz', d_test_paired)
        tasks.file_to_dict(tmp_prefix + '.unpaired.gz', d_test_unpaired)
        self.assertDictEqual(d_test_paired, d_correct_paired)
        self.assertDictEqual(d_test_unpaired, d_correct_unpaired)
        os.unlink(tmp_prefix + '.paired.gz')
        os.unlink(tmp_prefix + '.unpaired.gz')


class TestDeinterleave(unittest.TestCase):
    def test_deinterleave(self):
        '''deinterleave should deal with an interleaved file correctly'''
        tmp_1 = 'tmp.deinterleaved_1.fa'
        tmp_2 = 'tmp.deinterleaved_2.fa'
        tasks.deinterleave(os.path.join(data_dir, 'sequences_test_interleaved.fa'), tmp_1, tmp_2)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_deinterleaved_1.fa'), tmp_1))
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_deinterleaved_2.fa'), tmp_2))

        tasks.deinterleave(os.path.join(data_dir, 'sequences_test_interleaved.fq'), tmp_1, tmp_2, fasta_out=True)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_deinterleaved_1.fa'), tmp_1))
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_deinterleaved_2.fa'), tmp_2))

        with self.assertRaises(tasks.Error):
            tasks.deinterleave(os.path.join(data_dir, 'sequences_test_interleaved_bad.fa'), tmp_1, tmp_2)
        os.unlink(tmp_1)
        os.unlink(tmp_2)


class TestEnumerateNames(unittest.TestCase):
    def test_enumerate_names(self):
        '''Test enomereate_names works with all options'''
        outfile = 'tmp.enumerate_seqs.fa'
        rename_out = outfile + '.rename'
        tasks.enumerate_names(os.path.join(data_dir, 'sequences_test_enumerate_names.fa'), outfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_enumerate_names.fa.out.start.1'), outfile))
        tasks.enumerate_names(os.path.join(data_dir, 'sequences_test_enumerate_names.fa'), outfile, rename_file=rename_out)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_enumerate_names.fa.out.start.1'), outfile))
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_enumerate_names.fa.out.start.1.rename_file'), rename_out))
        tasks.enumerate_names(os.path.join(data_dir, 'sequences_test_enumerate_names.fa'), outfile, start_index=2)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_enumerate_names.fa.out.start.2'), outfile))
        tasks.enumerate_names(os.path.join(data_dir, 'sequences_test_enumerate_names.fa'), outfile, keep_illumina_suffix=True)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_enumerate_names.fa.out.keep_suffix'), outfile))

        tasks.enumerate_names(os.path.join(data_dir, 'sequences_test_enumerate_names.fa'), outfile, suffix='.SUFFIX')
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_enumerate_names.fa.out.add_suffix'), outfile, shallow=False))
        os.unlink(outfile)
        os.unlink(rename_out)


class TestExpandNucleotides(unittest.TestCase):
    def test_expand_nucleoties(self):
        '''Test expand_nucleotides'''
        tmp = 'tmp.expanded'
        fq_in = os.path.join(data_dir, 'tasks_test_expend_nucleotides.in.fq')
        fa_in = os.path.join(data_dir, 'tasks_test_expend_nucleotides.in.fa')
        fq_expected = os.path.join(data_dir, 'tasks_test_expend_nucleotides.out.fq')
        fa_expected = os.path.join(data_dir, 'tasks_test_expend_nucleotides.out.fa')
        tasks.expand_nucleotides(fq_in, tmp)
        self.assertTrue(filecmp.cmp(fq_expected, tmp, shallow=False))
        os.unlink(tmp)
        tasks.expand_nucleotides(fa_in, tmp)
        self.assertTrue(filecmp.cmp(fa_expected, tmp, shallow=False))
        os.unlink(tmp)


class TestExtendGaps(unittest.TestCase):
    def test_trim_contigs(self):
        '''Test that gap extension works'''
        outfile = 'tmp.gap_extend.fa'
        tasks.trim_contigs(os.path.join(data_dir, 'sequences_test_trim_contigs.fa'), outfile, trim=2)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_trim_contigs.fa.out'), outfile))
        os.unlink(outfile)


class TestFastqToMiraXml(unittest.TestCase):
    def test_fastaq_to_mira_xml(self):
        '''check that fastaq_to_mira_xml makes the correct xml file from a fastq file'''
        tmp = 'tmp.mira.xml'
        tasks.fastaq_to_mira_xml(os.path.join(data_dir, 'sequences_test_good_file.fq'), tmp)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_good_file_mira.xml'), tmp))
        os.unlink(tmp)


class TestFastaqToOrfsGFF(unittest.TestCase):
    def test_fastaq_to_orfs_gff(self):
        '''Test fastaq_to_orfs_gff'''
        outfile = 'tmp.orfs.gff'
        tasks.fastaq_to_orfs_gff(os.path.join(data_dir, 'sequences_test_orfs.fa'), outfile, min_length=120)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_orfs.gff'), outfile, shallow=False))
        os.unlink(outfile)


class TestFilter(unittest.TestCase):
    def test_length_filter(self):
        '''Check that filtering by length works as expected'''
        infile = os.path.join(data_dir, 'sequences_test_length_filter.fa')
        correct_files = [os.path.join(data_dir, 'sequences_test_length_filter.min-0.max-1.fa'),
                         os.path.join(data_dir, 'sequences_test_length_filter.min-0.max-inf.fa'),
                         os.path.join(data_dir, 'sequences_test_length_filter.min-4.max-4.fa')]
        cutoffs = [(0, 1), (0, float('inf')), (4, 4)]

        for i in range(len(cutoffs)):
            outfile = 'tmp.length_filter.fa'
            tasks.filter(infile, outfile, minlength=cutoffs[i][0], maxlength=cutoffs[i][1])
            self.assertTrue(filecmp.cmp(correct_files[i], outfile))
            os.unlink(outfile)

    def test_regex_filter(self):
        '''Check that filtering by name regex works as expected'''
        infile = os.path.join(data_dir, 'sequences_test_filter_by_regex.fa')
        correct_files = [os.path.join(data_dir, 'sequences_test_filter_by_regex.numeric.fa'),
                         os.path.join(data_dir, 'sequences_test_filter_by_regex.first-of-pair.fa'),
                         os.path.join(data_dir, 'sequences_test_filter_by_regex.first-char-a.fa')]
        regexes = ['^[0-9]+$', '/1$', '^a']

        for i in range(len(regexes)):
            outfile = 'tmp.regex_filter.fa'
            tasks.filter(infile, outfile, regex=regexes[i])
            self.assertTrue(filecmp.cmp(correct_files[i], outfile))
            os.unlink(outfile)

    def test_ids_from_file_filter(self):
        '''Test that can extract reads from a file of read names'''
        infile = os.path.join(data_dir, 'sequences_test_filter_by_ids_file.fa')
        outfile = 'tmp.ids_file_filter.fa'
        tasks.filter(infile, outfile, ids_file=infile + '.ids')
        self.assertTrue(filecmp.cmp(infile + '.filtered', outfile))
        os.unlink(outfile)

    def test_invert_filter(self):
        '''Test that inverting filtering works'''
        infile = os.path.join(data_dir, 'sequences_test_filter_by_ids_file.fa')
        outfile = 'tmp.ids_file_filter.fa'
        tasks.filter(infile, outfile, ids_file=infile + '.ids', invert=True)
        self.assertTrue(filecmp.cmp(infile + '.filtered.invert', outfile))
        os.unlink(outfile)


    def test_paired_both_pass(self):
        '''Test filter with paired file both pass'''
        infile1 = os.path.join(data_dir, 'tasks_test_filter_paired_both_pass.in_1.fa')
        infile2 = os.path.join(data_dir, 'tasks_test_filter_paired_both_pass.in_2.fa')
        outfile1 = 'tmp.filter_both_pass_1.fa'
        outfile2 = 'tmp.filter_both_pass_2.fa'
        expected1 = os.path.join(data_dir, 'tasks_test_filter_paired_both_pass.out_1.fa')
        expected2 = os.path.join(data_dir, 'tasks_test_filter_paired_both_pass.out_2.fa')
        tasks.filter(infile1, outfile1, mate_in=infile2, mate_out=outfile2, minlength=3)
        self.assertTrue(filecmp.cmp(outfile1, expected1, shallow=False))
        self.assertTrue(filecmp.cmp(outfile2, expected2, shallow=False))
        os.unlink(outfile1)
        os.unlink(outfile2)


    def test_paired_one_pass(self):
        '''Test filter with paired file one pass'''
        infile1 = os.path.join(data_dir, 'tasks_test_filter_paired_one_pass.in_1.fa')
        infile2 = os.path.join(data_dir, 'tasks_test_filter_paired_one_pass.in_2.fa')
        outfile1 = 'tmp.filter_one_pass_1.fa'
        outfile2 = 'tmp.filter_one_pass_2.fa'
        expected1 = os.path.join(data_dir, 'tasks_test_filter_paired_one_pass.out_1.fa')
        expected2 = os.path.join(data_dir, 'tasks_test_filter_paired_one_pass.out_2.fa')
        tasks.filter(infile1, outfile1, mate_in=infile2, mate_out=outfile2, both_mates_pass=False, minlength=3)
        self.assertTrue(filecmp.cmp(outfile1, expected1, shallow=False))
        self.assertTrue(filecmp.cmp(outfile2, expected2, shallow=False))
        os.unlink(outfile1)
        os.unlink(outfile2)


class TestGetSeqsFlankingGaps(unittest.TestCase):
    def test_get_seqs_flanking_gaps(self):
        outfile = 'tmp.seqs_flanking_gaps'
        tasks.get_seqs_flanking_gaps(os.path.join(data_dir, 'sequences_test_get_seqs_flanking_gaps.fa'), outfile, 3, 3)
        self.assertTrue(filecmp.cmp(outfile, os.path.join(data_dir, 'sequences_test_get_seqs_flanking_gaps.fa.out')))
        os.unlink(outfile)


class TestInterleave(unittest.TestCase):
    def test_interleave(self):
        '''Check that interleave works as expected'''
        tmp = 'tmp.interleaved.fa'
        tasks.interleave(os.path.join(data_dir, 'sequences_test_deinterleaved_1.fa'),
                         os.path.join(data_dir, 'sequences_test_deinterleaved_2.fa'),
                         tmp)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_interleaved.fa'), tmp))

        tasks.interleave(os.path.join(data_dir, 'sequences_test_deinterleaved_no_suffixes_1.fa'),
                         os.path.join(data_dir, 'sequences_test_deinterleaved_no_suffixes_2.fa'),
                         tmp, suffix1='/1', suffix2='/2')
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_interleaved_with_suffixes.fa'), tmp))

        with self.assertRaises(tasks.Error):
            tasks.interleave(os.path.join(data_dir, 'sequences_test_deinterleaved_bad_1.fa'),
                             os.path.join(data_dir, 'sequences_test_deinterleaved_bad_2.fa'),
                             tmp)

        with self.assertRaises(tasks.Error):
            tasks.interleave(os.path.join(data_dir, 'sequences_test_deinterleaved_bad2_1.fa'),
                             os.path.join(data_dir, 'sequences_test_deinterleaved_bad2_2.fa'),
                             tmp)
        os.unlink(tmp)


class TestMakeRandomContigs(unittest.TestCase):
    def test_make_random_contigs(self):
        '''Test make_random_contigs()'''
        # Can't guarantee same results from random (even using same seed), so
        # just check sequence names and lengths
        def files_are_equal(file1, file2):
            seqs1 = {}
            seqs2 = {}
            tasks.file_to_dict(file1, seqs1)
            tasks.file_to_dict(file2, seqs2)
            if len(seqs1) != len(seqs2):
                return False

            for name in seqs1:
                seq1 = seqs1[name]
                seq2 = seqs2[name]
                if seq1.id != seq2.id:
                    return False
                if len(seq1) != len(seq2):
                    return False

            return True

        tmp = 'tmp.random_contigs.fa'
        tasks.make_random_contigs(2, 3, tmp)
        self.assertTrue(files_are_equal(os.path.join(data_dir, 'sequences_test_make_random_contigs.default.fa'), tmp))
        tasks.make_random_contigs(2, 3, tmp, prefix='p')
        self.assertTrue(files_are_equal(os.path.join(data_dir, 'sequences_test_make_random_contigs.prefix-p.fa'), tmp))
        tasks.make_random_contigs(2, 3, tmp, first_number=42)
        self.assertTrue(files_are_equal(os.path.join(data_dir, 'sequences_test_make_random_contigs.first-42.fa'), tmp))
        tasks.make_random_contigs(28, 3, tmp, name_by_letters=True)
        self.assertTrue(files_are_equal(os.path.join(data_dir, 'sequences_test_make_random_contigs.name-by-letters.fa'), tmp))
        os.unlink(tmp)


class TestMeanLength(unittest.TestCase):
    def test_mean_length(self):
        '''Test mean_length'''
        expected = [3, 2, 3, 4, 4]
        limits = [1, 2, 3, 4, None]
        assert len(expected) == len(limits)
        for i in range(len(expected)):
            mean = tasks.mean_length(os.path.join(data_dir, 'tasks_test_mean_length.fa'), limit=limits[i])
            self.assertEqual(expected[i], mean)


class TestMergeToOneSeq(unittest.TestCase):
    def test_merge_to_one_seq_fa(self):
        '''Test merge_to_one_seq with fasta'''
        tmp = 'tmp.merged.fa'
        tasks.merge_to_one_seq(os.path.join(data_dir, 'sequences_test_merge_to_one_seq.fa'), tmp)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_merge_to_one_seq.merged.fa'), tmp, shallow=False))
        os.unlink(tmp)

    def test_merge_to_one_seq_fq(self):
        '''Test merge_to_one_seq with fastq'''
        tmp = 'tmp.merged.fq'
        tasks.merge_to_one_seq(os.path.join(data_dir, 'sequences_test_merge_to_one_seq.fq'), tmp)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_merge_to_one_seq.merged.fq'), tmp, shallow=False))
        os.unlink(tmp)

class TestReverseComplement(unittest.TestCase):
    def test_reverse_complement(self):
        '''reverse_complement should correctly reverse complement each seq in a file'''
        tmp = 'tmp.revcomp.fa'
        tasks.reverse_complement(os.path.join(data_dir, 'sequences_test.fa'), tmp)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_revcomp.fa'), tmp))
        os.unlink(tmp)


class TestScaffoldsToContigs(unittest.TestCase):
    def test_scaffolds_to_contigs(self):
        '''Test scaffolds_to_contigs'''
        tmp = 'tmp.contigs.fa'
        tasks.scaffolds_to_contigs(os.path.join(data_dir, 'utils_test_scaffolds.fa'), tmp)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'utils_test_scaffolds.fa.to_contigs.fa'), tmp))
        os.unlink(tmp)

    def test_scaffolds_to_contigs_number_contigs(self):
        '''Test scaffolds_to_contigs with contig numbering'''
        tmp = 'tmp.contigs.fa'
        tasks.scaffolds_to_contigs(os.path.join(data_dir, 'utils_test_scaffolds.fa'), tmp, number_contigs=True)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'utils_test_scaffolds.fa.to_contigs.number_contigs.fa'), tmp))
        os.unlink(tmp)


class TestSearchForSeq(unittest.TestCase):
    def test_search_for_seq(self):
        '''Test that sequence search finds all hits'''
        tmp = 'tmp.search.fa'
        tasks.search_for_seq(os.path.join(data_dir, 'sequences_test_search_string.fa'), tmp, 'AGA')
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_search_string.fa.hits'), tmp))
        os.unlink(tmp)


class TestSequenceTrim(unittest.TestCase):
    def test_sequence_trim(self):
        '''Test sequence_trim'''
        tmp1 = 'tmp.trimmed_1.fa'
        tmp2 = 'tmp.trimmed_2.fa'
        in1 = os.path.join(data_dir, 'tasks_test_sequence_trim_1.fa')
        in2 = os.path.join(data_dir, 'tasks_test_sequence_trim_2.fa')
        to_trim = os.path.join(data_dir, 'tasks_test_sequences_to_trim.fa')
        expected1 = os.path.join(data_dir, 'tasks_test_sequence_trim_1.trimmed.fa')
        expected2 = os.path.join(data_dir, 'tasks_test_sequence_trim_2.trimmed.fa')
        tasks.sequence_trim(in1, in2, tmp1, tmp2, to_trim, min_length=10, check_revcomp=True)
        self.assertTrue(filecmp.cmp(expected1, tmp1))
        self.assertTrue(filecmp.cmp(expected2, tmp2))
        os.unlink(tmp1)
        os.unlink(tmp2)


class ToFastg(unittest.TestCase):
    def test_to_fastg_ids_set(self):
        '''Test to_fastg when ids are a set'''
        infile = os.path.join(data_dir, 'tasks_test_to_fastg.fasta')
        tmpfile = 'tmp.to_fastg.fastg'
        expected = os.path.join(data_dir, 'tasks_test_to_fastg.fastg')
        ids = {'seq2'}
        tasks.to_fastg(infile, tmpfile, circular=ids)
        self.assertTrue(filecmp.cmp(expected, tmpfile, shallow=False))
        os.unlink(tmpfile)


    def test_to_fastg_ids_file(self):
        '''Test to_fastg when ids in a file'''
        infile = os.path.join(data_dir, 'tasks_test_to_fastg.fasta')
        tmpfile = 'tmp.to_fastg.fastg'
        expected = os.path.join(data_dir, 'tasks_test_to_fastg.fastg')
        ids_file = os.path.join(data_dir, 'tasks_test_to_fastg.ids_to_circularise')
        tasks.to_fastg(infile, tmpfile, circular=ids_file)
        self.assertTrue(filecmp.cmp(expected, tmpfile, shallow=False))
        os.unlink(tmpfile)


class TestTranslate(unittest.TestCase):
    def test_translate(self):
        '''Test translate works in each frame'''
        tmp = 'tmp.translated.fa'
        for i in range(3):
            tasks.translate(os.path.join(data_dir, 'sequences_test_translate.fa'), tmp, frame=i)
            self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_translate.fa.frame' + str(i)), tmp))

        os.unlink(tmp)


class TestTrim(unittest.TestCase):
    def test_trim(self):
        '''trim should correctly trim each seq in a file'''
        tmp = 'tmp.trim.fq'
        tasks.trim(os.path.join(data_dir, 'sequences_test_untrimmed.fq'), tmp, 2, 1)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_trimmed.fq'), tmp))
        os.unlink(tmp)


    def test_trim_Ns_at_end(self):
        '''Test Ns at ends of sequences trimmed OK'''
        tmp = 'tmp.trim.fa'
        tasks.trim_Ns_at_end(os.path.join(data_dir, 'sequences_test_trim_Ns_at_end.fa'), tmp)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_trim_Ns_at_end.fa.trimmed'), tmp))
        os.unlink(tmp)


class TestFileToDict(unittest.TestCase):
    def test_file_to_dict(self):
        '''check file_to_dict fills dictionary correctly'''
        d_test = {}
        d = {}
        tasks.file_to_dict(os.path.join(data_dir, 'sequences_test.fa'), d_test)
        for i in range(1,5):
            d[str(i)] = sequences.Fasta(str(i),'ACGTA')

        self.assertSequenceEqual(d_test.keys(),d.keys())
        for i in range(1,5):
            key = str(i)
            self.assertEqual(d_test[key].id, d[key].id)
            self.assertEqual(d_test[key].seq, d[key].seq)


class TestLengthsFromFai(unittest.TestCase):
    def test_lengths_from_fai(self):
        '''Check lengths_from_fai gets the length of each seq OK'''
        d = {}
        lengths = {str(x):x for x in range(1,5)}
        tasks.lengths_from_fai(os.path.join(data_dir, 'sequences_test_fai_test.fa.fai'), d)
        self.assertSequenceEqual(d.keys(), lengths.keys())
        for i in d:
            self.assertEqual(int(i), d[i])


class TestLengthOffsetsFromFai(unittest.TestCase):
    def test_length_offsets_from_fai(self):
        '''Test length_offsets_from_fai'''
        got = tasks.length_offsets_from_fai(os.path.join(data_dir, 'tasks_test_length_offsets_from_fai.fa.fai'))
        expected = {'seq1': 0, 'seq2': 42, 'seq3': 43}
        self.assertEqual(expected, got)


class TestSplit(unittest.TestCase):
    def test_split_by_base_count(self):
        '''Check that fasta/q files get split by base count correctly'''
        infile = os.path.join(data_dir, 'sequences_test_split_test.fa')
        outprefix = 'tmp.sequences_test_split_test.fa.test'
        length2files = {2: ['1','2','3','4'],
                        3: ['1','2','3'],
                        4: ['1', '2', '3'],
                        6: ['1', '2']}
        for l in length2files:
            tasks.split_by_base_count(infile, outprefix, l)
            for x in range(len(length2files[l])):
                file_index = str(length2files[l][x])
                fname = outprefix + '.' + file_index
                self.assertTrue(filecmp.cmp(fname, infile + '.' + str(l) + '.' + file_index))
                os.unlink(fname)

        # check that limiting the number of files works
        tasks.split_by_base_count(infile, outprefix, 6, 2)
        for i in range(1,4):
            test_file = outprefix + '.' + str(i)
            self.assertTrue(filecmp.cmp(test_file, os.path.join(data_dir, 'sequences_test_split_test.fa.6.limit2.') + str(i)))
            os.unlink(test_file)

        # check big sequence not broken
        tasks.split_by_base_count(os.path.join(data_dir, 'sequences_test_split_test.long.fa'), outprefix, 2)
        self.assertTrue(filecmp.cmp(outprefix + '.1', os.path.join(data_dir, 'sequences_test_split_test.long.fa.2.1')))
        self.assertTrue(filecmp.cmp(outprefix + '.2', os.path.join(data_dir, 'sequences_test_split_test.long.fa.2.2')))
        os.unlink(outprefix + '.1')
        os.unlink(outprefix + '.2')

    def test_split_by_fixed_size(self):
        '''Test fasta/q file split by fixed size'''
        infile = os.path.join(data_dir, 'sequences_test_split_fixed_size.fa')
        outprefix = 'tmp.sequences_test_split'
        tasks.split_by_fixed_size(infile, outprefix, 4, 1)

        for i in range(1,7,1):
            correct = os.path.join(data_dir, 'sequences_test_split_fixed_size.fa.split.' + str(i))
            test = outprefix + '.' + str(i)
            self.assertTrue(filecmp.cmp(test, correct))
            os.unlink(test)

        test_coords = outprefix + '.coords'
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_split_fixed_size.fa.split.coords'), test_coords))
        os.unlink(test_coords)

    def test_split_by_fixed_size_exclude_Ns(self):
        infile = os.path.join(data_dir, 'sequences_test_split_fixed_size.fa')
        outprefix = 'tmp.sequences_test_split'
        tasks.split_by_fixed_size(infile, outprefix, 4, 1, skip_if_all_Ns=True)

        for i in range(1,5,1):
            correct = os.path.join(data_dir, 'sequences_test_split_fixed_size.fa.split.skip_if_all_Ns.' + str(i))
            test = outprefix + '.' + str(i)
            self.assertTrue(filecmp.cmp(test, correct))
            os.unlink(test)

        test_coords = outprefix + '.coords'
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_split_fixed_size.fa.split.skip_if_all_Ns.coords'), test_coords))
        os.unlink(test_coords)

    def test_split_by_fixed_size_onefile(self):
        infile = os.path.join(data_dir, 'sequences_test_split_fixed_size_onefile.fa')
        tmp_out = 'tmp.sequences_test_split_fixed_size_onefile.fa'
        expected =  os.path.join(data_dir, 'sequences_test_split_fixed_size_onefile.out.fa')
        tasks.split_by_fixed_size_onefile(infile, tmp_out, chunk_size=3, tolerance=1)
        self.assertTrue(filecmp.cmp(expected, tmp_out))
        os.unlink(tmp_out)

    def test_split_by_fixed_size_onefile_exclude_Ns(self):
        infile = os.path.join(data_dir, 'sequences_test_split_fixed_size_onefile.fa')
        tmp_out = 'tmp.sequences_test_split_fixed_size_onefile.skip_Ns.fa'
        expected =  os.path.join(data_dir, 'sequences_test_split_fixed_size_onefile.skip_Ns.out.fa')
        tasks.split_by_fixed_size_onefile(infile, tmp_out, chunk_size=3, tolerance=1, skip_if_all_Ns=True)
        self.assertTrue(filecmp.cmp(expected, tmp_out))
        os.unlink(tmp_out)

class TestCountSequences(unittest.TestCase):
    def test_count_sequences(self):
        '''Check that count_sequences does as expected'''
        self.assertEqual(2, tasks.count_sequences(os.path.join(data_dir, 'sequences_test_good_file.fq')))
        self.assertEqual(4, tasks.count_sequences(os.path.join(data_dir, 'sequences_test.fa')))
        self.assertEqual(0, tasks.count_sequences(os.path.join(data_dir, 'sequences_test_empty_file')))

class TestGetIds(unittest.TestCase):
    def test_get_ids(self):
        '''Check that IDs extracted correctly from fasta/q file'''
        tmpfile = 'tmp.ids'
        tasks.get_ids(os.path.join(data_dir, 'sequences_test.fa'), tmpfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test.fa.ids'), tmpfile))
        os.unlink(tmpfile)


class TestFastaToFakeQual(unittest.TestCase):
    def test_fasta_to_fake_qual(self):
        '''Test fasta_to_fake_qual'''
        tmpfile = 'tmp.qual'
        infile = os.path.join(data_dir, 'tasks_test_fasta_to_fake_qual.in.fa')
        tasks.fastaq_to_fake_qual(infile, tmpfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'tasks_test_fasta_to_fake_qual.out.default.qual'), tmpfile, shallow=False))
        os.unlink(tmpfile)
        tasks.fastaq_to_fake_qual(infile, tmpfile, q=42)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'tasks_test_fasta_to_fake_qual.out.q42.qual'), tmpfile, shallow=False))
        os.unlink(tmpfile)


class TestFastaToFastq(unittest.TestCase):
    def test_fasta_to_fastq(self):
        '''Check fasta_to_fastq converts files as expected'''
        tasks.fasta_to_fastq(os.path.join(data_dir, 'sequences_test.fa'),
                             os.path.join(data_dir, 'sequences_test.fa.qual'),
                             'tmp.fq')
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test.fasta_to_fastq.fq'), 'tmp.fq'))

        with self.assertRaises(tasks.Error):
            tasks.fasta_to_fastq(os.path.join(data_dir, 'sequences_test.fa'),
                                 os.path.join(data_dir, 'sequences_test.fa.qual.bad'),
                                 'tmp.fq')

        os.unlink('tmp.fq')


class TestReplaceBases(unittest.TestCase):
    def test_sequences_replace_bases(self):
        '''Check that fasta file gets all bases replaced OK'''
        tmpfile = 'tmp.replace_bases.fa'
        tasks.replace_bases(os.path.join(data_dir, 'sequences_test_fastaq_replace_bases.fa'), tmpfile, 'T', 'X')
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_fastaq_replace_bases.expected.fa'), tmpfile))
        os.unlink(tmpfile)



class TestSortBySize(unittest.TestCase):
    def test_sort_by_size(self):
        '''Test sort_by_size'''
        infile = os.path.join(data_dir, 'tasks_test_sort_by_size.in.fa')
        tmpfile = 'tmp.sorted.fa'
        tasks.sort_by_size(infile, tmpfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'tasks_test_sort_by_size.out.fa'), tmpfile, shallow=False))
        tasks.sort_by_size(infile, tmpfile, smallest_first=True)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'tasks_test_sort_by_size.out.rev.fa'), tmpfile, shallow=False))
        os.unlink(tmpfile)

class TestSortByName(unittest.TestCase):
    def test_sort_by_name(self):
        '''Test sort_by_name'''
        infile = os.path.join(data_dir, 'tasks_test_sort_by_name.in.fa')
        tmpfile = 'tmp.sort_by_name.fa'
        tasks.sort_by_name(infile, tmpfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'tasks_test_sort_by_name.out.fa'), tmpfile, shallow=False))
        os.unlink(tmpfile)


class TestStripIlluminaSuffix(unittest.TestCase):
    def test_strip_illumina_suffix(self):
        '''Check illumina suffixes stripped correctly off read names'''
        tmpfile = 'tmp.stripped.fa'
        tasks.strip_illumina_suffix(os.path.join(data_dir, 'sequences_test_strip_illumina_suffix.fq'), tmpfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_strip_illumina_suffix.fq.stripped'), tmpfile))
        os.unlink(tmpfile)


class TestStatsFromFai(unittest.TestCase):
    def test_stats_from_fai_nonempty(self):
        '''Test task stats_from_fai non-empty file'''
        infile = os.path.join(data_dir, 'tasks_test_stats_from_fai.in.fai')
        got = tasks.stats_from_fai(infile)
        expected = {
            'longest': 10,
            'shortest': 1,
            'N50': 4,
            'mean': 4.2,
            'number': 5,
            'total_length': 21
        }
        self.assertEqual(expected, got)


    def test_stats_from_fai_empty(self):
        '''Test task stats_from_fai empty file'''
        infile = os.path.join(data_dir, 'tasks_test_stats_from_fai.in.empty.fai')
        got = tasks.stats_from_fai(infile)
        expected = {
            'longest': 0,
            'shortest': 0,
            'N50': 0,
            'mean': 0,
            'number': 0,
            'total_length': 0
        }
        self.assertEqual(expected, got)


class TestToBoulderio(unittest.TestCase):
    def test_to_boulderio(self):
        '''Test task to_boulderio'''
        tmpfile = 'tmp.boulder'
        tasks.to_boulderio(os.path.join(data_dir, 'tasks_test_to_boulderio.in.fa'), tmpfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'tasks_test_to_boulderio.out.boulder'), tmpfile, shallow=False))
        os.unlink(tmpfile)


class TestToFasta(unittest.TestCase):
    def test_to_fasta(self):
        '''Test to_fasta'''
        tmpfile = 'tmp.to_fasta'
        infiles = [
            'sequences_test_good_file.fq',
            'sequences_test_gffv3.gff',
            'sequences_test_gffv3.no_FASTA_line.gff',
            'sequences_test.embl',
            'sequences_test.gbk',
            'sequences_test_phylip.interleaved',
            'sequences_test_phylip.interleaved2',
            'sequences_test_phylip.sequential'
        ]
        infiles = [os.path.join(data_dir, x) for x in infiles]
        expected_outfiles = [x + '.to_fasta' for x in infiles]

        for i in range(len(infiles)):
            tasks.to_fasta(infiles[i], tmpfile)
            self.assertTrue(filecmp.cmp(expected_outfiles[i], tmpfile))

        tasks.to_fasta(os.path.join(data_dir, 'sequences_test.fa'), tmpfile, line_length=3)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test.line_length3.fa'), tmpfile))
        tasks.to_fasta(os.path.join(data_dir, 'sequences_test_strip_after_whitespace.fa'), tmpfile, strip_after_first_whitespace=True)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_strip_after_whitespace.fa.to_fasta'), tmpfile))
        os.unlink(tmpfile)

    def test_to_fasta_strip_after_whitespace_non_unique(self):
        '''Test strip_after_whitespace with non-unique names'''
        tmpfile = 'tmp.strip_after_whitespace.fa'
        infile = os.path.join(data_dir, 'sequences_test.to_fasta.strip_after_whitespace_non_unique.in.fa')
        expected = os.path.join(data_dir, 'sequences_test.to_fasta.strip_after_whitespace_non_unique.out.fa')

        with self.assertRaises(tasks.Error):
            tasks.to_fasta(infile, tmpfile, strip_after_first_whitespace=True, check_unique=True)

        tasks.to_fasta(infile, tmpfile, strip_after_first_whitespace=True, check_unique=False)
        self.assertTrue(filecmp.cmp(tmpfile, expected, shallow=False))
        os.unlink(tmpfile)

    def test_to_fasta_strip_after_whitespace_unique(self):
        '''Test strip_after_whitespace with unique names'''
        tmpfile = 'tmp.strip_after_whitespace.fa'
        infile = os.path.join(data_dir, 'sequences_test.to_fasta.strip_after_whitespace_unique.in.fa')
        expected = os.path.join(data_dir, 'sequences_test.to_fasta.strip_after_whitespace_unique.out.fa')
        tasks.to_fasta(infile, tmpfile, strip_after_first_whitespace=True, check_unique=True)
        self.assertTrue(filecmp.cmp(tmpfile, expected, shallow=False))
        os.unlink(tmpfile)

class TestToUniqueByID(unittest.TestCase):
    def test_to_unique_by_id(self):
        '''Test to_unique_by_id()'''
        tmpfile = 'tmp.unique_by_id.fa'
        tasks.to_unique_by_id(os.path.join(data_dir, 'sequences_test_to_unique_by_id.fa'), tmpfile)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_to_unique_by_id.fa.out'), tmpfile))
        os.unlink(tmpfile)


class TestToFastaUnion(unittest.TestCase):
    def test_to_fasta_union(self):
        '''Test to_fasta_union'''
        tmpfile = 'tmp.to_fasta_union'
        tasks.to_fasta_union(os.path.join(data_dir, 'sequences_test_to_fasta_union.in.fa'), tmpfile, seqname='testname')
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'sequences_test_to_fasta_union.out.fa'), tmpfile, shallow=False))
        os.unlink(tmpfile)


if __name__ == '__main__':
    unittest.main()




import argparse
from pyfastaq import tasks

def run(description):
    parser = argparse.ArgumentParser(
        description = 'Given a file of capillary reads, makes an interleaved file of read pairs (where more than read from same ligation, takes the longest read) and a file of unpaired reads. Replaces the .p1k/.q1k part of read names to denote fwd/rev reads with /1 and /2',
        usage = 'fastaq capillary_to_pairs <infile> <outfiles prefix>')
    parser.add_argument('infile', help='Name of input fasta/q file')
    parser.add_argument('outprefix', help='Prefix of output files', metavar='outfiles prefix')
    options = parser.parse_args()
    tasks.capillary_to_pairs(options.infile, options.outprefix)




import argparse
from pyfastaq import tasks

def run(description):
    parser = argparse.ArgumentParser(
        description = 'Splits a multi sequence file into separate files. Does not split sequences. Puts up to max_bases into each split file. The exception is that any sequence longer than max_bases is put into its own file.',
        usage = 'fastaq split_by_base_count [options] <infile> <outprefix> <max_bases>')
    parser.add_argument('infile', help='Name of input file to be split')
    parser.add_argument('outprefix', help='Name of output file')
    parser.add_argument('max_bases', type=int, help='Max bases in each output split file', metavar='max_bases')
    parser.add_argument('--max_seqs', type=int, help='Max number of sequences in each output split file [no limit]', metavar='INT')

    options = parser.parse_args()
    tasks.split_by_base_count(options.infile, options.outprefix, options.max_bases, options.max_seqs)



import argparse
import random
from math import floor, ceil
import sys
from pyfastaq import sequences, utils

def run(description):
    parser = argparse.ArgumentParser(
        description = 'Makes perfect paired end fastq reads from a sequence file, with insert sizes sampled from a normal distribution. Read orientation is innies. Output is an interleaved FASTQ file.',
        usage = 'fastaq to_perfect_reads [options] <infile> <outfile> <mean insert size> <insert std deviation> <mean coverage> <read length>')
    parser.add_argument('infile', help='Name of input file')
    parser.add_argument('outfile', help='Name of output file')
    parser.add_argument('mean_insert', type=int, help='Mean insert size of read pairs', metavar='mean insert size')
    parser.add_argument('insert_std', type=float, help='Standard devation of insert size', metavar='insert std deviation')
    parser.add_argument('coverage', type=float, help='Mean coverage of the reads', metavar='mean coverage')
    parser.add_argument('readlength', type=int, help='Length of each read', metavar='read length')
    parser.add_argument('--fragments', help='Write FASTA sequences of fragments (i.e. read pairs plus sequences in between them) to the given filename', metavar='FILENAME')
    parser.add_argument('--no_n', action='store_true', help='Don\'t allow any N or n characters in the reads')
    parser.add_argument('--seed', type=int, help='Seed for random number generator. Default is to use python\'s default', default=None, metavar='INT')
    options = parser.parse_args()

    random.seed(a=options.seed)

    seq_reader = sequences.file_reader(options.infile)
    fout = utils.open_file_write(options.outfile)
    pair_counter = 1

    if options.fragments:
        fout_frags = utils.open_file_write(options.fragments)

    for ref in seq_reader:
        # check if current seq is long enough
        if len(ref) < options.mean_insert + 4 * options.insert_std:
            print('Warning, sequence ', ref.id, ' too short.  Skipping it...', file=sys.stderr)
            continue

        # work out how many reads to simulate
        read_pairs = int(0.5 * options.coverage * len(ref) / options.readlength)

        # it's possible that we pick the same fragment twice, in which case the
        # reads would get the same name. So remember the frag coords
        used_fragments = {}  # (middle_position, length) => count

        # do the simulation:  pick insert size from normal distribution, and
        # position in genome from uniform distribution
        x = 0
        while x < read_pairs:
            isize = int(random.normalvariate(options.mean_insert, options.insert_std))
            while isize > len(ref) or isize < options.readlength:
                isize = int(random.normalvariate(options.mean_insert, options.insert_std))
            middle_pos = random.randint(ceil(0.5 *isize), floor(len(ref) - 0.5 * isize))
            read_start1 = int(middle_pos - ceil(0.5 * isize))
            read_start2 = read_start1 + isize - options.readlength

            readname = ':'.join([ref.id, str(pair_counter), str(read_start1+1), str(read_start2+1)])

            fragment = (middle_pos, isize)
            if fragment in used_fragments:
                used_fragments[fragment] += 1
                readname += '.dup.' + str(used_fragments[fragment])
            else:
                used_fragments[fragment] = 1

            read1 = sequences.Fastq(readname + '/1', ref.seq[read_start1:read_start1 + options.readlength], 'I' * options.readlength)
            read2 = sequences.Fastq(readname + '/2', ref.seq[read_start2:read_start2 + options.readlength], 'I' * options.readlength)


            if options.no_n and ('n' in read1.seq or 'N' in read1.seq or 'n' in read2.seq or 'N' in read2.seq):
                continue

            read2.revcomp()

            print(read1, file=fout)
            print(read2, file=fout)

            if options.fragments:
                frag = sequences.Fasta(readname, ref.seq[read_start1:read_start2 + options.readlength])
                print(frag, file=fout_frags)

            pair_counter += 1
            x += 1

    utils.close(fout)
    if options.fragments:
        utils.close(fout_frags)



import argparse
from pyfastaq import tasks

def run(description):
    parser = argparse.ArgumentParser(
        description = description,
        usage = 'fastaq replace_bases <infile> <outfile> <old> <new>')
    parser.add_argument('infile', help='Name of input file')
    parser.add_argument('outfile', help='Name of output file')
    parser.add_argument('old', help='Base to be replaced')
    parser.add_argument('new', help='Replace with this letter')
    options = parser.parse_args()
    tasks.replace_bases(options.infile, options.outfile, options.old, options.new)



import argparse
from pyfastaq import tasks

def run(description):
    parser = argparse.ArgumentParser(
        description = 'Splits a multi sequence file into separate files. Splits sequences into chunks of a fixed size. Aims for chunk_size chunks in each file, but allows a little extra, so chunk can be up to (chunk_size + tolerance), to prevent tiny chunks made from the ends of sequences',
        usage = 'fastaq chunker [options] <infile> <out> <chunk size> <tolerance>')
    parser.add_argument('infile', help='Name of input file to be split')
    parser.add_argument('out', help='Prefix of output file. If --onefile used, then name of single output file')
    parser.add_argument('chunk_size', type=int, help='Size of each chunk')
    parser.add_argument('tolerance', type=int, help='Tolerance allowed in chunk size')
    parser.add_argument('--onefile', action='store_true', help='Output all the sequences in one file')
    parser.add_argument('--skip_all_Ns', action='store_true', help='Do not output any sequence that consists of all Ns')
    options = parser.parse_args()
    if options.onefile:
        tasks.split_by_fixed_size_onefile(
            options.infile,
            options.out,
            options.chunk_size,
            options.tolerance,
            skip_if_all_Ns=options.skip_all_Ns
        )
    else:
        tasks.split_by_fixed_size(
            options.infile,
            options.out,
            options.chunk_size,
            options.tolerance,
            skip_if_all_Ns=options.skip_all_Ns
        )



import argparse
from pyfastaq import tasks

def run(description):
    parser = argparse.ArgumentParser(
        description = 'Writes a GFF file of open reading frames from a sequence file',
        usage = 'fastaq to_orfs_gff [options] <infile> <outfile>')
    parser.add_argument('--min_length', type=int, help='Minimum length of ORF, in nucleotides [%(default)s]', default=300, metavar='INT')
    parser.add_argument('infile', help='Name of input file')
    parser.add_argument('outfile', help='Name of output GFF file')
    options = parser.parse_args()
    tasks.fastaq_to_orfs_gff(options.infile, options.outfile, min_length=options.min_length)



import argparse
import sys
import random
from pyfastaq import sequences, utils, intervals

def run(description):
    parser = argparse.ArgumentParser(
        description = description,
        usage = 'fastaq add_indels [options] <infile> <outfile>')
    parser.add_argument('infile', help='Name of input file')
    parser.add_argument('outfile', help='Name of output file')
    parser.add_argument('-d','--delete', action='append', help='Delete the given bases from the given sequence. Format same as samtools view: name:start-end. This option can be used multiple times (once for each region to delete). Overlapping coords will be merged before deleting', metavar='Name:start:bases')
    parser.add_argument('--delete_range', help='Deletes bases starting at position P in each sequence of the input file. Deletes start + (n-1)*step bases from sequence n.', metavar='P,start,step')
    parser.add_argument('-i','--insert', action='append', help='Insert a random string of bases at the given position. Format is name:position:number_to_add. Bases are added after the position. This option can be used multiple times', metavar='Name:start:bases')
    parser.add_argument('--insert_range', help='Inserts random bases starting after position P in each sequence of the input file. Inserts start + (n-1)*step bases into sequence n.', metavar='P,start,step')
    options = parser.parse_args()

    test_ops = [int(x is not None) for x in [options.delete, options.insert, options.delete_range, options.insert_range]]

    if sum(test_ops) != 1:
        print('Must use one of --delete, --insert, --delete_range, --insert_range. Cannot continue', file=sys.stderr)
        sys.exit(1)


    def range2dic(range_in):
        if range_in is None:
            return {}
        (pos, start, step) = range_in.split(',')
        d = {}
        d['pos'] = int(pos) - 1
        d['bases'] = int(start)
        d['step'] = int(step)
        return d

    delete_range = range2dic(options.delete_range)
    insert_range = range2dic(options.insert_range)


    # convert the -d regions into sequence name, start and end coords
    to_delete = {}
    if options.delete:
        for s in options.delete:
            id, coords = s.rsplit(':')
            start, end = [int(x)-1 for x in coords.split('-')]
            if id not in to_delete:
                to_delete[id] = []
            to_delete[id].append(intervals.Interval(start, end))


    to_insert = {}
    if options.insert:
        for s in options.insert:
            id, pos, bases = s.rsplit(':',2)
            pos = int(pos) - 1
            bases = int(bases)
            if id not in to_insert:
                to_insert[id] = []
            to_insert[id].append((pos, bases))


    assert len(to_delete) * len(to_insert) == 0

    # merge overlapping regions to be deleted
    for l in to_delete.values():
        intervals.merge_overlapping_in_list(l)

    # sort positions to be inserted
    for l in to_insert.values():
        l.sort()

    # read in the fasta/q file and print outfile with deleted sequences
    seq_reader = sequences.file_reader(options.infile)
    f = utils.open_file_write(options.outfile)

    for seq in seq_reader:
        if seq.id in to_delete:
            # delete regions for this sequence, but start at the end so the
            # coords don't get messed up after the first deletion
            for inter in reversed(to_delete[seq.id]):
                seq.seq = seq.seq[:inter.start] + seq.seq[inter.end + 1:]
        elif options.delete_range:
            seq.seq = seq.seq[:delete_range['pos']] + seq.seq[delete_range['pos'] + delete_range['bases']:]
            delete_range['bases'] += delete_range['step']
        elif seq.id in to_insert:
            for pos, bases in reversed(to_insert[seq.id]):
                seq.seq = seq.seq[:pos + 1] + ''.join([random.choice('ACGT') for x in range(bases)]) + seq.seq[pos + 1:]
        elif options.insert_range:
            seq.seq = seq.seq[:insert_range['pos'] + 1] + ''.join([random.choice('ACGT') for x in range(insert_range['bases'])]) +  seq.seq[insert_range['pos'] + 1:]
            insert_range['bases'] += insert_range['step']

        print(seq, file=f)

    utils.close(f)



class Error (Exception): pass


class Interval:
    '''A class to deal with intervals in a genome. Can do things like intersections, unions etc'''
    def __init__(self, start, end):
        try:
            self.start = int(start)
            self.end = int(end)
        except ValueError:
            raise Error('Error making interval from :"' + str(start) + '" and "' + str(end) + '"')

        if self.end < self.start:
            raise Error('Error making interval ' + str(self) + '.  end < start.')

    def __len__(self):
        return self.end - self.start + 1

    def __eq__(self, other):
        return type(other) is type(self) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return '(' + str(self.start) + ',' + str(self.end) + ')'

    def __lt__(self, i):
        return self.start < i.start or (self.start == i.start and self.end < i.end)

    def __le__(self, i):
        return self.start < i.start or (self.start == i.start and self.end <= i.end)

    def distance_to_point(self, p):
        '''Returns the distance from the point to the interval. Zero if the point lies inside the interval.'''
        if self.start <= p <= self.end:
            return 0
        else:
            return min(abs(self.start - p), abs(self.end - p))

    def intersects(self, i):
        '''Returns true iff this interval intersects the interval i'''
        return self.start <= i.end and i.start <= self.end

    def contains(self, i):
        '''Returns true iff this interval contains the interval i'''
        return self.start <= i.start and i.end <= self.end

    def union(self, i):
        '''If intervals intersect, returns their union, otherwise returns None'''
        if self.intersects(i) or self.end + 1 == i.start or i.end + 1 == self.start:
            return Interval(min(self.start, i.start), max(self.end, i.end))
        else:
            return None

    def union_fill_gap(self, i):
        '''Like union, but ignores whether the two intervals intersect or not'''
        return Interval(min(self.start, i.start), max(self.end, i.end))

    def intersection(self, i):
        '''If intervals intersect, returns their intersection, otherwise returns None'''
        if self.intersects(i):
            return Interval(max(self.start, i.start), min(self.end, i.end))
        else:
            return None


def intersection(l1, l2):
    '''Returns intersection of two lists.  Assumes the lists are sorted by start positions'''
    if len(l1) == 0 or len(l2) == 0:
        return []

    out = []
    l2_pos = 0

    for l in l1:
        while l2_pos < len(l2) and l2[l2_pos].end < l.start:
            l2_pos += 1

        if l2_pos == len(l2):
            break

        while l2_pos < len(l2) and l.intersects(l2[l2_pos]):
            out.append(l.intersection(l2[l2_pos]))
            l2_pos += 1

        l2_pos = max(0, l2_pos - 1)

    return out


def merge_overlapping_in_list(l):
    '''Sorts list, merges any overlapping intervals, and also adjacent intervals. e.g.
       [0,1], [1,2] would be merge to [0,.2].'''
    i = 0
    l.sort()

    while i < len(l) - 1:
        u = l[i].union(l[i+1])
        if u is not None:
            l[i] = u
            l.pop(i+1)
        else:
            i += 1


def remove_contained_in_list(l):
    '''Sorts list in place, then removes any intervals that are completely
       contained inside another interval'''
    i = 0
    l.sort()

    while i < len(l) - 1:
       if l[i+1].contains(l[i]):
           l.pop(i)
       elif l[i].contains(l[i+1]):
           l.pop(i+1)
       else:
           i += 1


def length_sum_from_list(l):
    '''Returns total length of intervals from a list'''
    return sum([len(x) for x in l])



import argparse
from pyfastaq import tasks

def run(description):
    parser = argparse.ArgumentParser(
        description = 'Trims sequences off the start of all sequences in a pair of sequence files, whenever there is a perfect match. Only keeps a read pair if both reads of the pair are at least a minimum length after any trimming',
        usage = 'fastaq sequence_trim [options] <infile_1> <infile_2> <outfile_1> <outfile_2> <trim_seqs>')
    parser.add_argument('--min_length', type=int, help='Minimum length of output sequences [%(default)s]', default=50, metavar='INT')
    parser.add_argument('--revcomp', action='store_true', help='Trim the end of each sequence if it matches the reverse complement. This option is intended for PCR primer trimming') 
    parser.add_argument('infile_1', help='Name of forward fasta/q file to be trimmed')
    parser.add_argument('infile_2', help='Name of reverse fasta/q file to be trimmed')
    parser.add_argument('outfile_1', help='Name of output forward fasta/q file')
    parser.add_argument('outfile_2', help='Name of output reverse fasta/q file')
    parser.add_argument('trim_seqs', help='Name of file of sequences to search for at the start of each input sequence')
    options = parser.parse_args()
    tasks.sequence_trim(
        options.infile_1,
        options.infile_2,
        options.outfile_1,
        options.outfile_2,
        options.trim_seqs,
        min_length=options.min_length,
        check_revcomp=options.revcomp
    )



import argparse
from pyfastaq import tasks

def run(description):
    parser = argparse.ArgumentParser(
        description = 'Filters a sequence file by sequence length and/or by name matching a regular expression',
        usage = 'fastaq filter [options] <infile> <outfile>')
    parser.add_argument('--min_length', type=int, help='Minimum length of sequence to keep [%(default)s]', default=0, metavar='INT')
    parser.add_argument('--max_length', type=float, help='Maximum length of sequence to keep [%(default)s]', default=float('inf'), metavar='INT')
    parser.add_argument('--regex', help='If given, only reads with a name matching the regular expression will be kept')
    parser.add_argument('--ids_file', help='If given, only reads whose ID is in th given file will be used. One ID per line of file.', metavar='FILENAME')
    parser.add_argument('-v', '--invert', action='store_true', help='Only keep sequences that do not match the filters')

    mate_group = parser.add_argument_group('Mate file for read pairs options')
    mate_group.add_argument('--mate_in', help='Name of mates input file. If used, must also provide --mate_out', metavar='FILENAME')
    mate_group.add_argument('--mate_out', help='Name of mates output file', metavar='FILENAME')
    mate_group.add_argument('--both_mates_pass', action='store_true', help='By default, if either mate passes filter, then both reads output. Use this flag to require that both reads of a pair pass the filter')

    parser.add_argument('infile', help='Name of input file to be filtered')
    parser.add_argument('outfile', help='Name of output file')
    options = parser.parse_args()
    tasks.filter(options.infile,
                 options.outfile,
                 minlength=options.min_length,
                 maxlength=options.max_length,
                 regex=options.regex,
                 ids_file=options.ids_file,
                 invert=options.invert,
                 mate_in=options.mate_in,
                 mate_out=options.mate_out,
                 both_mates_pass=options.both_mates_pass,
    )
